@@ -3,7 +3,7 @@
  * tftp.c
  *    main client file.
  *
- * $Id: tftp.c,v 1.32.2.2 2003/06/20 02:09:26 jp Exp $
+ * $Id: tftp.c,v 1.47 2004/03/15 23:55:56 jp Exp $
  *
  * Copyright (c) 2000 Jean-Pierre Lefebvre <helix@step.polymtl.ca>
  *                and Remi Lefebvre <remi@debian.org>
@@ -26,14 +26,22 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 
 #include <signal.h> 
 
+#if HAVE_READLINE
 #include <readline/readline.h>
 #include <readline/history.h>
+#endif
+
+#if HAVE_ARGZ
 #include <argz.h>
+#else
+#include "argz.h"
+#endif
 
 #include "tftp.h"
 #include "tftp_io.h"
@@ -43,21 +51,18 @@
 
 /* Maximum number of args on a line. 20 should be more than enough. */
 #define MAXARG  20
-
+/* for readline */
 #define HISTORY_FILE ".atftp_history"
 
-/*
- * General options. Used in tftp.c and tftp_file.c only.
- */
-int trace = 0;                  /* debugging information */
-int verbose = 0;                /* to print message at each step */
-
+/* defined as extern in tftp_file.c and mtftp_file.c, set by the signal
+   handler */
 int tftp_cancel = 0;
 
 /* local flags */
 int interactive = 1;            /* if false, we run in batch mode */
 int tftp_result = OK;           /* status of tftp_send_file or
-                                   tftp_receive_file */
+                                   tftp_receive_file, used for status() */
+
 /* Structure to hold some information that must be passed to
  * functions.
  */
@@ -66,10 +71,12 @@ struct client_data data;
 /* tftp.c local only functions. */
 static void signal_handler(int signal);
 int read_cmd(void);
-#if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
+#if HAVE_READLINE
+# if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
 int getc_func(FILE *fp);
 char **completion(const char *text, int start, int end);
 char *command_generator(const char *text, int state);
+# endif
 #endif
 void make_arg(char *string, int *argc, char ***argv);
 int process_cmd(int argc, char **argv);
@@ -82,6 +89,9 @@ int set_mode(int argc, char **argv);
 int set_option(int argc, char **argv);
 int put_file(int argc, char **argv);
 int get_file(int argc, char **argv);
+#ifdef HAVE_MTFTP
+int mtftp_opt(int argc, char **argv);
+#endif
 int quit(int argc, char **argv);
 int set_verbose(int argc, char **argv);
 int set_trace(int argc, char **argv);
@@ -100,6 +110,10 @@ struct command {
      {"option", set_option, "set RFC1350 options"},
      {"put", put_file, "upload a file to the host"},
      {"get", get_file, "download a file from the host"},
+#ifdef HAVE_MTFTP
+     {"mtftp", mtftp_opt, "set mtftp variables"},
+     {"mget", get_file, "download file from mtftp server"},
+#endif
      {"quit", quit, "exit tftp"},
      {"verbose", set_verbose, "toggle verbose mode"},
      {"trace", set_trace, "toggle trace mode"},
@@ -110,10 +124,6 @@ struct command {
      {NULL, NULL, NULL}
 };
 
-/* Defined in tftp_file.c and only called in this file. */
-int tftp_receive_file(struct client_data *data);
-int tftp_send_file(struct client_data *data);
-
 /*
  * Open the socket, register signal handler and
  * pass the control to read_cmd.
@@ -150,15 +160,20 @@ int main(int argc, char **argv)
      memcpy(data.tftp_options_reply, tftp_default_options,
             sizeof(tftp_default_options));
 
-     /* get the server entry */
-     data.sp = getservbyname("tftp", "udp");
-     if (data.sp == 0) {
-          fprintf(stderr, "tftp: udp/tftp, unknown service.\n");
-          exit(ERR);
-     }
-
-     /* default timeout value */
+     /* default options  */
+#ifdef HAVE_MTFTP
+     data.mtftp_client_port = 76;
+     Strncpy(data.mtftp_mcast_ip, "0.0.0.0", MAXLEN);
+     data.mtftp_listen_delay = 2;
+     data.mtftp_timeout_delay = 2;
+#endif
      data.timeout = TIMEOUT;
+     data.checkport = 1;
+     data.trace = 0;
+     data.verbose = 0;
+#ifdef DEBUG
+     data.delay = 0;
+#endif
 
      /* register SIGINT -> C-c */
      signal(SIGINT, signal_handler);
@@ -172,6 +187,10 @@ int main(int argc, char **argv)
      return OK;
 }
 
+/* 
+ * When we receive a signal, we set tftp_cancel in order to
+ * abort ongoing transfer.
+ */
 void signal_handler(int signal)
 {
      /* 
@@ -182,59 +201,87 @@ void signal_handler(int signal)
 }
 
 /*
- * Read command with a nice prompt and history :)
+ * Read commands with a nice prompt and history (if compile with
+ * libreadline. Otherway we only get the basic.
  */
 int read_cmd(void)
 {
      int run = 1;
+#if HAVE_READLINE
      char *string = NULL;
+#else
+     char string[MAXLEN];
+#endif
      int argc;
      char **argv = NULL;
+#if HAVE_READLINE
      char history[MAXLEN];
 
      if (getenv("HOME") != NULL)
           snprintf(history, sizeof(history), "%s/%s", getenv("HOME"), HISTORY_FILE);
      else
-          snprintf(history, sizeof(history), "/%s", HISTORY_FILE);
+          snprintf(history, sizeof(history), "%s", HISTORY_FILE);
 
-#if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
+# if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
      rl_attempted_completion_function = completion;
      rl_getc_function = getc_func;
+# endif
 #endif
 
+#if HAVE_READLINE
      using_history();
      read_history(history);
+#endif
 
      while (run)
      {
+#if HAVE_READLINE
           if ((string = readline("tftp> ")) == NULL)
           {
                fprintf(stderr, "\n");
                break;
           }
+#else
+          fprintf(stderr, "tftp> ");
+          if (fgets(string, MAXLEN, stdin) == NULL)
+          {
+               fprintf(stderr, "\n");
+               break;
+          }
+#endif
           else
           {
+#ifndef HAVE_READLINE
+               string[strlen(string)-1] = 0;
+#endif
                if (strlen(string) != 0)
                {
                     make_arg(string, &argc, &argv);
                     if (argc > 0)
                     {
+#if HAVE_READLINE
                          add_history(string);
+#endif
                          if (process_cmd(argc, argv) == QUIT)
                               run = 0;
                     }
+#if HAVE_READLINE
                     free(string);
+#endif
                }
           }
      }
 
+#if HAVE_READLINE
      /* save history */
      write_history(history);
+#endif
 
      return 0;
 }
 
-#if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
+#if HAVE_READLINE
+# if (HAVE_RL_COMPLETION_MATCHES | HAVE_COMPLETION_MATCHES)
 int getc_func(FILE *fp)
 {
      fd_set rfds;
@@ -298,6 +345,7 @@ char *command_generator(const char *text, int state)
      /* If no names matched, then return NULL. */
      return NULL;
 }
+# endif
 #endif
 
 /*
@@ -360,7 +408,8 @@ int process_cmd(int argc, char **argv)
  */
 int set_peer(int argc, char **argv)
 {
-     struct hostent *host;
+     struct hostent *host;      /* for host name lookup */
+     struct servent *sp;        /* server entry for tftp service */
 
      /* sanity check */
      if ((argc < 2) || (argc > 3))
@@ -368,6 +417,14 @@ int set_peer(int argc, char **argv)
           fprintf(stderr, "Usage: %s host-name [port]\n", argv[0]);
           return ERR;
      }
+
+     /* get the server entry */
+     sp = getservbyname("tftp", "udp");
+     if (sp == 0) {
+          fprintf(stderr, "tftp: udp/tftp, unknown service.\n");
+          return ERR;
+     }
+
      /* look up the host */
      host = gethostbyname(argv[1]);
      /* if valid, update s_inn structure */
@@ -377,10 +434,10 @@ int set_peer(int argc, char **argv)
           if (host->h_length > sizeof(data.sa_peer.sin_addr))
                host->h_length = sizeof(data.sa_peer.sin_addr);
           memcpy(&data.sa_peer.sin_addr, host->h_addr, host->h_length);
-          strncpy(data.hostname, host->h_name,
+          Strncpy(data.hostname, host->h_name,
                   sizeof(data.hostname));
           data.hostname[sizeof(data.hostname)-1] = 0;
-          data.sa_peer.sin_port = data.sp->s_port;
+          data.sa_peer.sin_port = sp->s_port;
      } 
      else
      {
@@ -391,17 +448,17 @@ int set_peer(int argc, char **argv)
      /* get the server port */
      if (argc == 3)
      {
-          data.sp->s_port = htons(atoi(argv[2]));
-          if (data.sp->s_port < 0)
+          sp->s_port = htons(atoi(argv[2]));
+          if (sp->s_port < 0)
           {
                fprintf(stderr, "%s: bad port number.\n", argv[2]);
                data.connected = 0;
                return ERR;
           }
-          data.sa_peer.sin_port = data.sp->s_port;
+          data.sa_peer.sin_port = sp->s_port;
      }
      /* copy port number to data structure */
-     data.port = ntohs(data.sp->s_port);
+     data.port = ntohs(sp->s_port);
 
      data.connected = 1;
      return OK;
@@ -424,10 +481,10 @@ int set_mode(int argc, char **argv)
           return OK;
      }
      if (strcasecmp("netascii", argv[1]) == 0)
-          strncpy(data.tftp_options[OPT_MODE].value, "netascii",
+          Strncpy(data.tftp_options[OPT_MODE].value, "netascii",
                   VAL_SIZE);
      else if (strcasecmp("octet", argv[1]) == 0)
-          strncpy(data.tftp_options[OPT_MODE].value, "octet",
+          Strncpy(data.tftp_options[OPT_MODE].value, "octet",
                   VAL_SIZE);
      else
      {
@@ -450,6 +507,24 @@ int set_option(int argc, char **argv)
      {
           fprintf(stderr, "Usage: option <option name> [option value]\n");
           fprintf(stderr, "       option disable <option name>\n");
+          if (data.tftp_options[OPT_TSIZE].specified)
+               fprintf(stderr, "  tsize:     enabled\n");
+          else
+               fprintf(stderr, "  tsize:     disabled\n");
+          if (data.tftp_options[OPT_BLKSIZE].specified)
+               fprintf(stderr, "  blksize:   %s\n",
+                       data.tftp_options[OPT_BLKSIZE].value);
+          else
+               fprintf(stderr, "  blksize:   disabled\n");
+          if (data.tftp_options[OPT_TIMEOUT].specified)
+               fprintf(stderr, "  timeout:   %s\n",
+                       data.tftp_options[OPT_TIMEOUT].value);
+          else
+               fprintf(stderr, "  timeout:   disabled\n");
+          if (data.tftp_options[OPT_MULTICAST].specified)
+               fprintf(stderr, "  multicast: enabled\n");
+          else
+               fprintf(stderr, "  multicast: disabled\n");
           return ERR;
      }
      /* if disabling an option */
@@ -517,13 +592,13 @@ int put_file(int argc, char **argv)
      }
      if (argc == 2)
      {
-          strncpy(data.local_file, argv[1], VAL_SIZE);
-          strncpy(data.tftp_options[OPT_FILENAME].value, argv[1], VAL_SIZE);
+          Strncpy(data.local_file, argv[1], VAL_SIZE);
+          Strncpy(data.tftp_options[OPT_FILENAME].value, argv[1], VAL_SIZE);
      }
      else
      {
-          strncpy(data.local_file, argv[1], VAL_SIZE);
-          strncpy(data.tftp_options[OPT_FILENAME].value, argv[2], VAL_SIZE);
+          Strncpy(data.local_file, argv[1], VAL_SIZE);
+          Strncpy(data.tftp_options[OPT_FILENAME].value, argv[2], VAL_SIZE);
      }
      
      /* open a UDP socket */
@@ -554,13 +629,28 @@ int put_file(int argc, char **argv)
  */
 int get_file(int argc, char **argv)
 {
+#if HAVE_READLINE
      char *string;
+#else
+     char string[MAXLEN];
+#endif
+#ifdef HAVE_MTFTP
+     int use_mtftp;
+#endif
+     char *tmp;
      FILE *fp;
      socklen_t len = sizeof(data.sa_local);
 
+#ifdef HAVE_MTFTP
+     if (strcmp(argv[0], "mget") == 0)
+          use_mtftp = 1;
+     else
+          use_mtftp = 0;
+#endif
+
      if ((argc < 2) || (argc > 3))
      {
-          fprintf(stderr, "Usage: get remote_file [local_file]\n");
+          fprintf(stderr, "Usage: %s remote_file [local_file]\n", argv[0]);
           return ERR;
      }
      if (!data.connected)
@@ -570,19 +660,19 @@ int get_file(int argc, char **argv)
      }
      if (argc == 2)
      {
-          strncpy(data.tftp_options[OPT_FILENAME].value,
+          Strncpy(data.tftp_options[OPT_FILENAME].value,
                   argv[1], VAL_SIZE);
-          string = strrchr(argv[1], '/');
-          if (string < argv[1])
-               string = argv[1];
+          tmp = strrchr(argv[1], '/');
+          if (tmp < argv[1])
+               tmp = argv[1];
           else
-               string++;
-          strncpy(data.local_file, string, VAL_SIZE);
+               tmp++;
+          Strncpy(data.local_file, tmp, VAL_SIZE);
      }
      else
      {
-          strncpy(data.local_file, argv[2], VAL_SIZE);
-          strncpy(data.tftp_options[OPT_FILENAME].value, argv[1], VAL_SIZE);
+          Strncpy(data.local_file, argv[2], VAL_SIZE);
+          Strncpy(data.tftp_options[OPT_FILENAME].value, argv[1], VAL_SIZE);
      }
 
      /* if interractive, verify if localfile exists */
@@ -594,13 +684,23 @@ int get_file(int argc, char **argv)
                if ((fp = fopen(data.local_file, "r")) != NULL)
                {
                     fclose(fp);
+#if HAVE_READLINE
                     string = readline("Overwite local file [y/n]? ");
+#else
+                    fprintf(stderr, "Overwite local file [y/n]? ");
+                    fgets(string, MAXLEN, stdin);
+                    string[strlen(string) - 1] = 0;
+#endif
                     if (!(strcasecmp(string, "y") == 0))
                     {
+#if HAVE_READLINE
                          free(string);
+#endif
                          return OK;
                     }
+#if HAVE_READLINE
                     free(string);
+#endif
                }
           }
      }
@@ -618,7 +718,13 @@ int get_file(int argc, char **argv)
 
      /* do the transfer */
      gettimeofday(&data.start_time, NULL);
-     tftp_result = tftp_receive_file(&data);
+#ifdef HAVE_MTFTP
+     if (use_mtftp)
+          tftp_result = tftp_mtftp_receive_file(&data);
+     else
+#endif
+          tftp_result = tftp_receive_file(&data);
+
      gettimeofday(&data.end_time, NULL);
 
      /* close the socket */
@@ -628,6 +734,57 @@ int get_file(int argc, char **argv)
      return OK;
 }
 
+#ifdef HAVE_MTFTP
+/* 
+ * Set ot get mtftp variable value
+ */
+int mtftp_opt(int argc, char **argv)
+{
+     if (argc != 3)
+     {
+          fprintf(stderr, "Usage: mtftp <option name> <option value>\n");
+          /* print current value of variables */
+          fprintf(stderr, "  client-port:   %d\n", data.mtftp_client_port);
+          fprintf(stderr, "  mcast-ip:      %s\n", data.mtftp_mcast_ip);
+          fprintf(stderr, "  listen-delay:  %d\n", data.mtftp_listen_delay);
+          fprintf(stderr, "  timeout-delay: %d\n", data.mtftp_timeout_delay);
+     }
+     else
+     {
+          if (strcmp(argv[1], "client-port") == 0)
+          {
+               data.mtftp_client_port = atoi(argv[2]);
+               fprintf(stderr, "mtftp client-port = %d\n",
+                       data.mtftp_client_port);
+          }
+          else if (strcmp(argv[1], "mcast-ip") == 0)
+          {
+               Strncpy(data.mtftp_mcast_ip, argv[2], MAXLEN);
+               fprintf(stderr, "mtftp mcast-ip = %s\n", data.mtftp_mcast_ip);
+          }
+          else if (strcmp(argv[1], "listen-delay") == 0)
+          {
+               data.mtftp_listen_delay = atoi(argv[2]);
+               fprintf(stderr, "mtftp listen-delay = %d\n",
+                       data.mtftp_listen_delay);
+          }
+          else if (strcmp(argv[1], "timeout-delay") == 0)
+          {
+               data.mtftp_timeout_delay = atoi(argv[2]);
+               fprintf(stderr, "mtftp timeout-delay = %d\n",
+                       data.mtftp_timeout_delay);
+          }
+          else
+          {
+               fprintf(stderr, "no mtftp variable named %s\n", argv[1]);
+               return ERR;
+          }
+     }
+     return OK;
+}
+
+#endif
+
 /*
  * Exit tftp
  */
@@ -638,14 +795,14 @@ int quit(int argc, char **argv)
 
 int set_verbose(int argc, char **argv)
 {
-     if (verbose)
+     if (data.verbose)
      {
-          verbose = 0;
+          data.verbose = 0;
           fprintf(stderr, "Verbose mode off.\n");
      }
      else
      {
-          verbose = 1;
+          data.verbose = 1;
           fprintf(stderr, "Verbose mode on.\n");
      }
      return OK;
@@ -653,14 +810,14 @@ int set_verbose(int argc, char **argv)
 
 int set_trace(int argc, char **argv)
 {
-     if (trace)
+     if (data.trace)
      {
-          trace = 0;
+          data.trace = 0;
           fprintf(stderr, "Trace mode off.\n");
      }
      else
      {
-          trace = 1;
+          data.trace = 1;
           fprintf(stderr, "Trace mode on.\n");
      }
      return OK;
@@ -669,7 +826,9 @@ int set_trace(int argc, char **argv)
 int status(int argc, char **argv)
 {
      struct timeval tmp;
-     HIST_ENTRY *history = previous_history();
+#if HAVE_READLINE
+     HIST_ENTRY *history = history_get(history_length-1);
+#endif
      char string[MAXLEN];
 
      timeval_diff(&tmp, &data.end_time, &data.start_time);
@@ -680,11 +839,11 @@ int status(int argc, char **argv)
           fprintf(stderr, "Connected:  %s port %d\n", data.hostname,
                   data.port);
      fprintf(stderr, "Mode:       %s\n", data.tftp_options[OPT_MODE].value);
-     if (verbose)
+     if (data.verbose)
           fprintf(stderr, "Verbose:    on\n");
      else
           fprintf(stderr, "Verbose:    off\n");
-     if (trace)
+     if (data.trace)
           fprintf(stderr, "Trace:      on\n");
      else
           fprintf(stderr, "Trace:      off\n");
@@ -705,11 +864,19 @@ int status(int argc, char **argv)
           fprintf(stderr, " multicast: enabled\n");
      else
           fprintf(stderr, " multicast: disabled\n");
-
+#ifdef HAVE_MTFTP
+     fprintf(stderr, "mtftp variables\n");
+     fprintf(stderr, " client-port:   %d\n", data.mtftp_client_port);
+     fprintf(stderr, " mcast-ip:      %s\n", data.mtftp_mcast_ip);
+     fprintf(stderr, " listen-delay:  %d\n", data.mtftp_listen_delay);
+     fprintf(stderr, " timeout-delay: %d\n", data.mtftp_timeout_delay);
+#endif
+#if HAVE_READLINE
      if (history)
           fprintf(stderr, "Last command: %s\n", history->line);
      else
           fprintf(stderr, "Last command: %s\n", "---");
+#endif
 
      if (strlen(data.tftp_options[OPT_FILENAME].value) > 0)
      {
@@ -717,14 +884,14 @@ int status(int argc, char **argv)
                   data.tftp_options[OPT_FILENAME].value);
           if (tftp_result == OK)
           {
-               print_eng((double)data.file_size, string, "%3.3f%cB");
+               print_eng((double)data.file_size, string, sizeof(string), "%3.3f%cB");
                fprintf(stderr, "  Bytes transfered:  %s\n", string);
                fprintf(stderr, "  Time of transfer: %8.3fs\n",
                        (double)(tmp.tv_sec + tmp.tv_usec * 1e-6));
                fprintf(stderr, "  Throughput:        ");
                print_eng((data.file_size /
                           (double)(tmp.tv_sec + tmp.tv_usec * 1e-6)),
-                         string, "%3.3f%cB/s\n");
+                         string, sizeof(string), "%3.3f%cB/s\n");
                fprintf(stderr, "%s", string);
           }
           else
@@ -775,8 +942,9 @@ int help(int argc, char **argv)
      return OK;
 }
 
-#define PUT 1
-#define GET 2
+#define PUT  1
+#define GET  2
+#define MGET 3
 
 /*
  * If tftp is called with --help, usage is printed and we exit.
@@ -796,73 +964,43 @@ int tftp_cmd_line_options(int argc, char **argv)
 
      int option_index = 0;
      static struct option options[] = {
-          { "help", 0, NULL, 'h' },
-          { "trace", 0, NULL, 'd'},
-          { "verbose", 0, NULL, 'v'},
+          { "get", 0, NULL, 'g'},
+#ifdef HAVE_MTFTP
+          { "mget", 0, NULL, 'G'},
+#endif
+          { "put", 0, NULL, 'p'},
+          { "local-file", 1, NULL, 'l'},
+          { "remote-file", 1, NULL, 'r'},
           { "tftp-timeout", 1, NULL, 'T'},
+          { "mode", 1, NULL, 'M'},
+          { "option", 1, NULL, 'O'},
+#if 1
           { "timeout", 1, NULL, 't'},
           { "blksize", 1, NULL, 'b'},
           { "tsize", 0, NULL, 's'},
           { "multicast", 0, NULL, 'm'},
-          { "local-file", 1, NULL, 'l'},
-          { "remote-file", 1, NULL, 'r'},
-          { "get", 0, NULL, 'g'},
-          { "put", 0, NULL, 'p'},
+#endif
+          { "mtftp", 1, NULL, '1'},
+          { "no-source-port-checking", 0, NULL, '0'},
+          { "verbose", 0, NULL, 'v'},
+          { "trace", 0, NULL, 'd'},
+#if DEBUG
+          { "delay", 1, NULL, 'D'},
+#endif
           { "version", 0, NULL, 'V'},
+          { "help", 0, NULL, 'h' },
           { 0, 0, 0, 0 }
      };
 
-     while ((c = getopt_long(argc, argv, "hdvT:t:b:sml:r:gpV",
+     /* Support old argument until 0.8 */
+     while ((c = getopt_long(argc, argv, /*"gpl:r:Vh"*/ "gpl:r:Vht:b:sm",
                              options, &option_index)) != EOF)
      {
           switch (c)
           {
-          case 'd':
-               snprintf(string, sizeof(string), "trace on");
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);            
-               break;
-          case 'v':
-               snprintf(string, sizeof(string), "verbose on");
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);            
-               break;
-          case 'T':
-               snprintf(string, sizeof(string), "timeout %s", optarg);
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);
-               break;
-          case 't':
-               snprintf(string, sizeof(string), "option timeout %s", optarg);
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);
-               break;
-          case 'b':
-               snprintf(string, sizeof(string), "option blksize %s", optarg);
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);
-               break;
-          case 's':
-               snprintf(string, sizeof(string), "option tsize");
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);
-               break;
-          case 'm':
-               snprintf(string, sizeof(string), "option multicast");
-               make_arg(string, &ac, &av);
-               process_cmd(ac, av);
-               break;
-          case 'l':
-               interactive = 0;
-               strncpy(local_file, optarg, MAXLEN);
-               break;
-          case 'r':
-               interactive = 0;
-               strncpy(remote_file, optarg, MAXLEN);
-               break;
           case 'g':
                interactive = 0;
-               if (action == PUT)
+               if ((action == PUT) || (action == MGET))
                {
                     fprintf(stderr, "two actions specified!\n");
                     exit(ERR);
@@ -870,9 +1008,19 @@ int tftp_cmd_line_options(int argc, char **argv)
                else
                     action = GET;
                break;
+          case 'G':
+               interactive = 0;
+               if ((action == PUT) || (action == GET))
+               {
+                    fprintf(stderr, "two actions specified!\n");
+                    exit(ERR);
+               }
+               else
+                    action = MGET;
+               break;
           case 'p':
                interactive = 0;
-               if (action == GET)
+               if ((action == GET) || (action == MGET))
                {
                     fprintf(stderr, "two actions specified!\n");
                     exit(ERR);
@@ -880,6 +1028,79 @@ int tftp_cmd_line_options(int argc, char **argv)
                else
                     action = PUT;
                break;
+          case 'l':
+               interactive = 0;
+               Strncpy(local_file, optarg, MAXLEN);
+               break;
+          case 'r':
+               interactive = 0;
+               Strncpy(remote_file, optarg, MAXLEN);
+               break;
+          case 'T':
+               snprintf(string, sizeof(string), "timeout %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case 'M':
+               snprintf(string, sizeof(string), "mode %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case 'O':
+               snprintf(string, sizeof(string), "option %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case '1':
+               snprintf(string, sizeof(string), "mtftp %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+#if 1
+          case 't':
+               fprintf(stderr, "--timeout deprecate, use --option instead\n");
+               snprintf(string, sizeof(string), "option timeout %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case 'b':
+               fprintf(stderr, "--blksize deprecate, use --option instead\n");
+               snprintf(string, sizeof(string), "option blksize %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case 's':
+               fprintf(stderr, "--tsize deprecate, use --option instead\n");
+               snprintf(string, sizeof(string), "option tsize");
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+          case 'm':
+               fprintf(stderr,
+                       "--multicast deprecate, use --option instead\n");
+               snprintf(string, sizeof(string), "option multicast");
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
+#endif
+          case '0':
+               data.checkport = 0;
+               break;
+          case 'v':
+               snprintf(string, sizeof(string), "verbose on");
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);            
+               break;
+          case 'd':
+               snprintf(string, sizeof(string), "trace on");
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);            
+               break;
+#if DEBUG
+          case 'D':
+               data.delay = atoi(optarg);
+               break;
+#endif
           case 'V':
                fprintf(stderr, "atftp-%s (client)\n", VERSION);
                exit(OK);
@@ -908,7 +1129,7 @@ int tftp_cmd_line_options(int argc, char **argv)
                snprintf(string, sizeof(string), "connect %s", argv[optind]);
           else
                snprintf(string, sizeof(string), "connect %s %s", argv[optind],
-                        argv[optind+1]);
+                                       argv[optind+1]);
           make_arg(string, &ac, &av);
           process_cmd(ac, av);
      }
@@ -921,13 +1142,17 @@ int tftp_cmd_line_options(int argc, char **argv)
           else if (action == GET)
                snprintf(string, sizeof(string), "get %s %s", remote_file,
                         local_file);
+          else if (action == MGET)
+               snprintf(string, sizeof(string), "mget %s %s", remote_file,
+                        local_file);
           else
           {
                fprintf(stderr, "No action specified in batch mode!\n");
                exit(ERR);
           }
           make_arg(string, &ac, &av);
-          process_cmd(ac, av);
+          if (process_cmd(ac, av) == ERR)
+               exit(ERR);
      }
      return OK;
 }
@@ -937,21 +1162,33 @@ void tftp_usage(void)
      fprintf(stderr,
              "Usage: tftp [options] [host] [port]\n"
              " [options] may be:\n"
-             "  -v, --verbose              : set verbose on\n"
-             "  -d, --debug                : set trace on\n"
-             "  -T, --tftp-timeout <value> : delay before"
-             " retransmission, client side)\n"
-             "  -t, --timeout <value>      : delay before"
-             " retransmition, server side (RFC2349)\n"
-             "  -b, --blocksize <value>    : specify blksize (RFC2348)\n"
-             "  -s, --tsize                : use 'tsize' from RFC2349\n"
-             "  -m, --multicast            : use 'multicast' from RFC2090\n"
-             "  -l, --local-file <file>    : local file to use\n"
-             "  -r, --remote-file <file>   : remote file name\n"
-             "  -g, --get                  : get file\n"
-             "  -p, --put                  : put file\n"
-             "  -h, --help                 : print this help\n"
-             "  -V, --version              : print version information\n"
+             "  -g, --get                : get file\n"
+#ifdef HAVE_MTFTP
+             "      --mget               : get file using mtftp\n"
+#endif
+             "  -p, --put                : put file\n"
+             "  -l, --local-file <file>  : local file name\n"
+             "  -r, --remote-file <file> : remote file name\n"
+             "  --tftp-timeout <value>   : delay before retransmission, client side\n"
+#if 0
+             "  t, --timeout <value>      : delay before retransmission, "
+                                           "server side (RFC2349)\n"
+             "  b, --blocksize <value>    : specify blksize to use (RFC2348)\n"
+             "  s, --tsize                : use 'tsize' (RFC2349)\n"
+             "  m, --multicast            : use 'multicast' (RFC2090)\n"
+#endif
+             "  --option <\"name value\">  : set option name to value\n"
+#ifdef HAVE_MTFTP
+             "  --mtftp <\"name value\">   : set mtftp variable to value\n"
+#endif
+             "  --no-source-port-checking: violate RFC, see man page\n"
+             "  --verbose                : set verbose mode on\n"
+             "  --trace                  : set trace mode on\n"
+#if DEBUG
+             "  --delay                  : add delay in state machine for debugging\n"
+#endif
+             "  -V, --version            : print version information\n"
+             "  -h, --help               : print this help\n"
              "\n"
              " [host] is the tftp server name\n"
              " [port] is the port to use\n" 
